/*
 * *****************************************************************************
 *
 * Copyright 2018 Gavin D. Howard
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * *****************************************************************************
 *
 * The entry point for bc.
 *
 */

#define _GNU_SOURCE 1

#include <stdarg.h>
#include <limits.h>
#include <stdint.h>
#include <stdbool.h>
#include <unistd.h>
#include <inttypes.h>
#include <string.h>
#include <stdio.h> /* for FILE */
#include <sys/types.h> /* size_t */
#include <sys/stat.h>

#include <ctype.h>
//#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
//#include <paths.h>
//#include <stdio.h>
#include <stdlib.h>
//#include <stdarg.h>
//#include <stddef.h>
//#include <string.h>
#include <poll.h>
//#include <sys/ioctl.h>
//#include <sys/mman.h>
//#include <sys/resource.h>
//#include <sys/socket.h>
//#include <sys/time.h>
//#include <sys/types.h>
//#include <sys/wait.h>
//#include <termios.h>
//#include <time.h>
//#include <sys/param.h>
//#include <pwd.h>
//#include <grp.h>
#include <locale.h>
#include <getopt.h>


#include "bc_interface.h"


#define ENABLE_FEATURE_BC_SIGNALS 1
#define IF_FEATURE_BC_SIGNALS(...) __VA_ARGS__
#define IF_NOT_FEATURE_BC_SIGNALS(...)

#define ENABLE_FEATURE_CLEAN_UP 0
#define IF_FEATURE_CLEAN_UP(...)
#define IF_NOT_FEATURE_CLEAN_UP(...) __VA_ARGS__

#define ENABLE_BC 1
#define IF_BC(...) __VA_ARGS__
#define IF_NOT_BC(...)

#define ENABLE_DC 1
#define IF_DC(...) __VA_ARGS__
#define IF_NOT_DC(...)

#define FAST_FUNC

#define ALIGN1

#define UNUSED_PARAM

#define MAIN_EXTERNALLY_VISIBLE

#define NORETURN __attribute__ ((__noreturn__))

#define NOINLINE __attribute__((__noinline__))

#define ALWAYS_INLINE __attribute__ ((always_inline)) inline

#define ARRAY_SIZE(x) ((unsigned)(sizeof(x) / sizeof((x)[0])))

typedef int smallint;

/* This struct is deliberately not defined. */
/* See docs/keep_data_small.txt */
struct globals;
/* '*const' ptr makes gcc optimize code much better.
 * Magic prevents ptr_to_globals from going into rodata.
 * If you want to assign a value, use SET_PTR_TO_GLOBALS(x) */
struct globals *const ptr_to_globals;
/* At least gcc 3.4.6 on mipsel system needs optimization barrier */
#define barrier() __asm__ __volatile__("":::"memory")
#define SET_PTR_TO_GLOBALS(x) do { \
	(*(struct globals**)&ptr_to_globals) = (void*)(x); \
	barrier(); \
} while (0)
#define FREE_PTR_TO_GLOBALS() do { \
	if (ENABLE_FEATURE_CLEAN_UP) { \
		free(ptr_to_globals); \
	} \
} while (0)

uint32_t option_mask32;

const int const_int_0;

const char *applet_name;

smallint xfunc_error_retval = 1;

smallint bb_got_signal;

const char bb_hexdigits_upcase[] ALIGN1 = "0123456789ABCDEF";

#define BB_VER "1.2.3"

#define GETOPT_RESET() (optind = 0)

void record_signo(int signo)
{
	bb_got_signal = signo;
}

void FAST_FUNC xfunc_die(void)
{
	exit(xfunc_error_retval);
}

ssize_t FAST_FUNC safe_read(int fd, void *buf, size_t count)
{
	ssize_t n;

	do {
		n = read(fd, buf, count);
	} while (n < 0 && errno == EINTR);

	return n;
}

ssize_t FAST_FUNC full_read(int fd, void *buf, size_t len)
{
	ssize_t cc;
	ssize_t total;

	total = 0;

	while (len) {
		cc = safe_read(fd, buf, len);

		if (cc < 0) {
			if (total) {
				/* we already have some! */
				/* user can do another read to know the error code */
				return total;
			}
			return cc; /* read() returns -1 on failure. */
		}
		if (cc == 0)
			break;
		buf = ((char *)buf) + cc;
		total += cc;
		len -= cc;
	}

	return total;
}

ssize_t FAST_FUNC safe_write(int fd, const void *buf, size_t count)
{
	ssize_t n;

	for (;;) {
		n = write(fd, buf, count);
		if (n >= 0 || errno != EINTR)
			break;
		/* Some callers set errno=0, are upset when they see EINTR.
		 * Returning EINTR is wrong since we retry write(),
		 * the "error" was transient.
		 */
		errno = 0;
		/* repeat the write() */
	}

	return n;
}

/*
 * Write all of the supplied buffer out to a file.
 * This does multiple writes as necessary.
 * Returns the amount written, or -1 on an error.
 */
ssize_t FAST_FUNC full_write(int fd, const void *buf, size_t len)
{
	ssize_t cc;
	ssize_t total;

	total = 0;

	while (len) {
		cc = safe_write(fd, buf, len);

		if (cc < 0) {
			if (total) {
				/* we already wrote some! */
				/* user can do another write to know the error code */
				return total;
			}
			return cc;  /* write() returns -1 on failure. */
		}

		total += cc;
		buf = ((const char *)buf) + cc;
		len -= cc;
	}

	return total;
}

void bb_verror_msg(const char *s, va_list p, const char* strerr) FAST_FUNC;

void FAST_FUNC bb_error_msg_and_die(const char *s, ...)
{
	va_list p;

	va_start(p, s);
	bb_verror_msg(s, p, NULL);
	va_end(p);
	xfunc_die();
}

void FAST_FUNC bb_error_msg(const char *s, ...)
{
	va_list p;

	va_start(p, s);
	bb_verror_msg(s, p, NULL);
	va_end(p);
}

void FAST_FUNC bb_perror_msg(const char *s, ...)
{
	va_list p;

	va_start(p, s);
	/* Guard against "<error message>: Success" */
	bb_verror_msg(s, p, errno ? strerror(errno) : NULL);
	va_end(p);
}

void FAST_FUNC bb_die_memory_exhausted(void)
{
	bb_error_msg_and_die("out of memory");
}

static int FAST_FUNC fflush_all(void)
{
	return fflush(NULL);
}

int FAST_FUNC bb_putchar(int ch)
{
	return putchar(ch);
}
char* FAST_FUNC xasprintf(const char *format, ...)
{
	va_list p;
	int r;
	char *string_ptr;

	va_start(p, format);
	r = vasprintf(&string_ptr, format, p);
	va_end(p);

	if (r < 0)
		bb_die_memory_exhausted();
	return string_ptr;
}

#define msg_eol "\n"

void FAST_FUNC bb_verror_msg(const char *s, va_list p, const char* strerr)
{
	char *msg, *msg1;
	char stack_msg[80];
	int applet_len, strerr_len, msgeol_len, used;

	if (!s) /* nomsg[_and_die] uses NULL fmt */
		s = ""; /* some libc don't like printf(NULL) */

	applet_len = strlen(applet_name) + 2; /* "applet: " */
	strerr_len = strerr ? strlen(strerr) : 0;
	msgeol_len = strlen(msg_eol);

	/* This costs ~90 bytes of code, but avoids costly
	 * malloc()[in vasprintf]+realloc()+memmove()+free() in 99% of cases.
	 * ~40% speedup.
	 */
	if ((int)sizeof(stack_msg) - applet_len > 0) {
		va_list p2;

		/* It is not portable to use va_list twice, need to va_copy it */
		va_copy(p2, p);
		used = vsnprintf(stack_msg + applet_len, (int)sizeof(stack_msg) - applet_len, s, p2);
		va_end(p2);
		msg = stack_msg;
		used += applet_len;
		if (used < (int)sizeof(stack_msg) - 3 - msgeol_len - strerr_len)
			goto add_pfx_and_sfx;
	}

	used = vasprintf(&msg, s, p);
	if (used < 0)
		return;

	/* This is ugly and costs +60 bytes compared to multiple
	 * fprintf's, but is guaranteed to do a single write.
	 * This is needed for e.g. httpd logging, when multiple
	 * children can produce log messages simultaneously. */

	/* can't use xrealloc: it calls error_msg on failure,
	 * that may result in a recursion */
	/* +3 is for ": " before strerr and for terminating NUL */
	msg1 = realloc(msg, applet_len + used + strerr_len + msgeol_len + 3);
	if (!msg1) {
		msg[used++] = '\n'; /* overwrites NUL */
		applet_len = 0;
	} else {
		msg = msg1;
		/* TODO: maybe use writev instead of memmoving? Need full_writev? */
		memmove(msg + applet_len, msg, used);
		used += applet_len;
 add_pfx_and_sfx:
		strcpy(msg, applet_name);
		msg[applet_len - 2] = ':';
		msg[applet_len - 1] = ' ';
		if (strerr) {
			if (s[0]) { /* not perror_nomsg? */
				msg[used++] = ':';
				msg[used++] = ' ';
			}
			strcpy(&msg[used], strerr);
			used += strerr_len;
		}
		strcpy(&msg[used], msg_eol);
		used += msgeol_len;
	}

	fflush_all();
	full_write(STDERR_FILENO, msg, used);
	if (msg != stack_msg)
		free(msg);
}

void FAST_FUNC bb_perror_msg_and_die(const char *s, ...)
{
	va_list p;

	va_start(p, s);
	/* Guard against "<error message>: Success" */
	bb_verror_msg(s, p, errno ? strerror(errno) : NULL);
	va_end(p);
	xfunc_die();
}

char* FAST_FUNC xstrdup(const char *s)
{
	char *t;

	if (s == NULL)
		return NULL;

	t = strdup(s);

	if (t == NULL)
		bb_die_memory_exhausted();

	return t;
}

void* FAST_FUNC xmalloc(size_t size)
{
	void *ptr = malloc(size);
	if (ptr == NULL && size != 0)
		bb_die_memory_exhausted();
	return ptr;
}

void* FAST_FUNC xzalloc(size_t size)
{
	void *ptr = xmalloc(size);
	memset(ptr, 0, size);
	return ptr;
}

void* FAST_FUNC xrealloc(void *ptr, size_t size)
{
	ptr = realloc(ptr, size);
	if (ptr == NULL && size != 0)
		bb_die_memory_exhausted();
	return ptr;
}

int FAST_FUNC safe_poll(struct pollfd *ufds, nfds_t nfds, int timeout)
{
	while (1) {
		int n = poll(ufds, nfds, timeout);
		if (n >= 0)
			return n;
		/* Make sure we inch towards completion */
		if (timeout > 0)
			timeout--;
		/* E.g. strace causes poll to return this */
		if (errno == EINTR)
			continue;
		/* Kernel is very low on memory. Retry. */
		/* I doubt many callers would handle this correctly! */
		if (errno == ENOMEM)
			continue;
		bb_perror_msg("poll");
		return n;
	}
}

ssize_t FAST_FUNC nonblock_immune_read(int fd, void *buf, size_t count)
{
	struct pollfd pfd[1];
	ssize_t n;

	while (1) {
		n = safe_read(fd, buf, count);
		if (n >= 0 || errno != EAGAIN)
			return n;
		/* fd is in O_NONBLOCK mode. Wait using poll and repeat */
		pfd[0].fd = fd;
		pfd[0].events = POLLIN;
		/* note: safe_poll pulls in printf */
		safe_poll(pfd, 1, -1);
	}
}

void* FAST_FUNC xmalloc_read(int fd, size_t *maxsz_p)
{
	char *buf;
	size_t size, rd_size, total;
	size_t to_read;
	struct stat st;

	to_read = maxsz_p ? *maxsz_p : (INT_MAX - 4095); /* max to read */

	/* Estimate file size */
	st.st_size = 0; /* in case fstat fails, assume 0 */
	fstat(fd, &st);
	/* /proc/N/stat files report st_size 0 */
	/* In order to make such files readable, we add small const */
	size = (st.st_size | 0x3ff) + 1;

	total = 0;
	buf = NULL;
	while (1) {
		if (to_read < size)
			size = to_read;
		buf = xrealloc(buf, total + size + 1);
		rd_size = full_read(fd, buf + total, size);
		if ((ssize_t)rd_size == (ssize_t)(-1)) { /* error */
			free(buf);
			return NULL;
		}
		total += rd_size;
		if (rd_size < size) /* EOF */
			break;
		if (to_read <= rd_size)
			break;
		to_read -= rd_size;
		/* grow by 1/8, but in [1k..64k] bounds */
		size = ((total / 8) | 0x3ff) + 1;
		if (size > 64*1024)
			size = 64*1024;
	}
	buf = xrealloc(buf, total + 1);
	buf[total] = '\0';

	if (maxsz_p)
		*maxsz_p = total;
	return buf;
}

void* FAST_FUNC xmalloc_open_read_close(const char *filename, size_t *maxsz_p)
{
	char *buf;
	int fd;

	fd = open(filename, O_RDONLY);
	if (fd < 0)
		return NULL;

	buf = xmalloc_read(fd, maxsz_p);
	close(fd);
	return buf;
}

void* FAST_FUNC xmalloc_xopen_read_close(const char *filename, size_t *maxsz_p)
{
	void *buf = xmalloc_open_read_close(filename, maxsz_p);
	if (!buf)
		bb_perror_msg_and_die("can't read '%s'", filename);
	return buf;
}

char* FAST_FUNC skip_whitespace(const char *s)
{
	/* In POSIX/C locale (the only locale we care about: do we REALLY want
	 * to allow Unicode whitespace in, say, .conf files? nuts!)
	 * isspace is only these chars: "\t\n\v\f\r" and space.
	 * "\t\n\v\f\r" happen to have ASCII codes 9,10,11,12,13.
	 * Use that.
	 */
	while (*s == ' ' || (unsigned char)(*s - 9) <= (13 - 9))
		s++;

	return (char *) s;
}

char* FAST_FUNC skip_non_whitespace(const char *s)
{
	while (*s != '\0' && *s != ' ' && (unsigned char)(*s - 9) > (13 - 9))
		s++;

	return (char *) s;
}

int FAST_FUNC sigaction_set(int signum, const struct sigaction *act)
{
	return sigaction(signum, act, NULL);
}

void FAST_FUNC signal_SA_RESTART_empty_mask(int sig, void (*handler)(int))
{
	struct sigaction sa;
	memset(&sa, 0, sizeof(sa));
	/*sigemptyset(&sa.sa_mask);*/
	sa.sa_flags = SA_RESTART;
	sa.sa_handler = handler;
	sigaction_set(sig, &sa);
}

static const struct option bc_args_lopt[] = {

	{ "expression", required_argument, NULL, 'e' },
	{ "file", required_argument, NULL, 'f' },
	{ "help", no_argument, NULL, 'h' },
	{ "interactive", no_argument, NULL, 'i' },
	{ "mathlib", no_argument, NULL, 'l' },
	{ "quiet", no_argument, NULL, 'q' },
	{ "standard", no_argument, NULL, 's' },
	{ "version", no_argument, NULL, 'v' },
	{ "warn", no_argument, NULL, 'w' },
	{ "extended-register", no_argument, NULL, 'x' },
	{ 0, 0, 0, 0 },

};

//static const char* const bc_args_opt = "e:f:hilqsvVwx";
static const char* const bc_args_opt = "hilqsvVwx";

//void bc_args_exprs(BcVec *exprs, const char *str) {
//	bc_vec_concat(exprs, str);
//	bc_vec_concat(exprs, "\n");
//}

//BcStatus bc_args_file(BcVec *exprs, const char *file) {
//
//	BcStatus s;
//	char *buf;
//
//	s = bc_read_file(file, &buf);
//	if (s) return s;
//
//	bc_args_exprs(exprs, buf);
//	free(buf);
//
//	return s;
//}

unsigned FAST_FUNC string_array_len(char **argv)
{
	char **start = argv;

	while (*argv)
		argv++;

	return argv - start;
}

static char** ext_bc_args(char **argv) //, BcVec *exs)
{
	int c, argc;

	argc = string_array_len(argv);
	optind = 0;
	while ((c = getopt_long(argc, argv, bc_args_opt, bc_args_lopt, NULL)) != -1) {

		switch (c) {

			case 0:
			{
				// This is the case when a long option is found.
				break;
			}

//			case 'e':
//			{
//				bc_args_exprs(exs, optarg);
//				break;
//			}
//
//			case 'f':
//			{
//				s = bc_args_file(exs, optarg);
//				break;
//			}
//
			case 'h':
			{
//				bc_vm_info(bcg.help);
				exit(0);
				break;
			}

#if BC_ENABLED
			case 'i':
			{
//				if (!BC_IS_BC) s = BC_STATUS_INVALID_OPTION;
				option_mask32 |= BC_FLAG_I;
				break;
			}

			case 'l':
			{
//				if (!BC_IS_BC) s = BC_STATUS_INVALID_OPTION;
				option_mask32 |= BC_FLAG_L;
				break;
			}

			case 'q':
			{
//				if (!BC_IS_BC) s = BC_STATUS_INVALID_OPTION;
				option_mask32 |= BC_FLAG_Q;
				break;
			}

			case 's':
			{
//				if (!BC_IS_BC) s = BC_STATUS_INVALID_OPTION;
				option_mask32 |= BC_FLAG_S;
				break;
			}

			case 'w':
			{
//				if (!BC_IS_BC) s = BC_STATUS_INVALID_OPTION;
				option_mask32 |= BC_FLAG_W;
				break;
			}
#endif // BC_ENABLED

			case 'V':
			case 'v':
			{
//				bc_vm_info(NULL);
				exit(0);
				break;
			}

//#if DC_ENABLED
//			case 'x':
//			{
//				if (BC_IS_BC) s = BC_STATUS_INVALID_OPTION;
//				option_mask32 |= BC_FLAG_X;
//				break;
//			}
//#endif // DC_ENABLED

			// Getopt printed an error message, but we should exit.
			case '?':
			default:
			{
				exit(1); //BC_STATUS_INVALID_OPTION);
				break;
			}
		}
	}

//	if (exs->len > 1 || !BC_IS_BC) option_mask32 |= BC_FLAG_Q;

	return argv + optind;
}

static void bb_show_usage(void)
{
	printf("HELP GOES HERE\n");
	exit(0);
}

static unsigned str2num(const char *arg)
{
	unsigned long v;
	char *end;
	if (!isalnum(arg[0])) {
		errno = EINVAL;
		return 0;
	}
	errno = 0;
        v = strtoul(arg, &end, 10);
	if (v > UINT_MAX) {
		errno = ERANGE;
	}
	if (*end) {
		errno = EINVAL;
	}
	return v;
}


#include "bc_code.h"


int main(int argc, char *argv[]) {

	int s;

	setlocale(LC_ALL, "");

#if !DC_ENABLED
	s = bc_main(argc, argv);
#elif !BC_ENABLED
	s = dc_main(argc, argv);
#else
	char *name = strrchr(argv[0], '/');

	if (!strcmp(name, "dc")) {
		applet_name = "dc";
		s = dc_main(argc, argv);
	} else {
		applet_name = "bc";
		s = bc_main(argc, argv);
	}
#endif

	return s;
}
